/* Version 1.5 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <math.h>
#include <unistd.h>
#include <float.h>
#include <limits.h>
#include <signal.h>
/* Vienna */
#include <ViennaRNA/fold.h>
#include <ViennaRNA/RNAstruct.h>
#include <ViennaRNA/dist_vars.h>
#include <ViennaRNA/treedist.h>
#include <ViennaRNA/utils.h>


#define MAXLEN (200)
#define BUFSIZE (100)
 
/* Base */
#define A (0)
#define C (1)
#define G (2)
#define U (3)

/* target & initial */

char targetStrG[MAXLEN]="";
char initSeqG[MAXLEN]="";
int distanceG = 1; /* 1 or 2 */
int rnalenG = 200;

/* functions */
void Init(void);
void Distance1(void);
void Distance2(void);
char RNAi2c(int);
int RNAc2i(char);
int Args(int,char**);


int main(int argc, char **argv)
{
  Args(argc,argv);
  Init();

  if(distanceG==1)
    Distance1();
  else if(distanceG==2)
    Distance2();
  else{
    fprintf(stderr,"main: Error\n");
    exit(1);
  }
  return 0;
}

/****************/
/* Initlization */
/****************/
void Init(void)
{

  /* Examine the arguments. This is after Args() is called */
  if(initSeqG[0]=='\0'){
    fprintf(stderr,"Init: No initial sequence!");
    exit(1);
  }
  else if(targetStrG[0]=='\0'){
    fprintf(stderr,"Init: No target structure!");
    exit(1);
  }
  else if(strlen(initSeqG)!=strlen(targetStrG)){
    fprintf(stderr,"Init: The length of the initial sequence must be the same as that of the target structure\n");
    exit(1);
  }

  /* other initilization */
  initialize_fold(MAXLEN);
  rnalenG = strlen(initSeqG);
}


 

/***************************
 *  Molecule Annotation    *
 ***************************/

/* fold all possible one step mutant, and print its 2nd structure
   and the structural distance to the target structure and the
   free energy */
void Distance1(void)
{
  int i,j;
  char sequence[MAXLEN];
  char structure[MAXLEN];
  int bp_dist;
  double energy;

  strcpy(sequence,initSeqG);
  for(i=0;i<rnalenG;i++){
    for(j=1;j<=3;j++){
      /* mutate */
      sequence[i]=RNAi2c((RNAc2i(initSeqG[i])+j)%4);

      /* Fold */
      energy = (double) fold(sequence,structure);

      bp_dist = bp_distance(structure,targetStrG);

      /* Output */
      printf("%s %s %d %f\n",sequence,structure,bp_dist,energy);
    }
    /* bring mutation back to original */
    sequence[i]=initSeqG[i];
  }
  fflush(stdout);
}

void Distance2(void)
{
  int i,j;
  int ci,cj;
  char sequence[MAXLEN];
  char structure[MAXLEN];
  int bp_dist;
  double energy;

  strcpy(sequence,initSeqG);

  for(i=0;i<rnalenG;i++)
    for(j=i+1;j<rnalenG;j++){
      for(ci=1;ci<=3;ci++) 
	for(cj=1;cj<=3;cj++){
	  /* mutate */
	  sequence[i]=RNAi2c((RNAc2i(initSeqG[i])+ci)%4);
	  sequence[j]=RNAi2c((RNAc2i(initSeqG[j])+cj)%4);
	  
	  /* Fold */
	  energy = (double) fold(sequence,structure);
	  bp_dist = bp_distance(structure,targetStrG);
	  
	  /* Output */
	  printf("%s %s %d %f\n",sequence,structure,bp_dist,energy);
	}
      /* bring mutation back to original */
      sequence[i]=initSeqG[i];
      sequence[j]=initSeqG[j];
    }
  fflush(stdout);
}


/*********************
 * RNAi2c c2i etc.   *
 *********************/

/* char RNAi2c(int rnaint)
   Parameter: give a RNA digit in 'rnaint' to change it to RNA character.
   Description: RNA ind to character.
   Return value: 'A' for 0, 'C' for 1, 'G' for 2, 'U' for '3', '\0' for error. 
   Note: Better to use macro may be. 
*/
char RNAi2c(int rnaint)
{
  switch (rnaint) {
  case 0: 
    return ('A');
  case 1:
    return ('C');
  case 2:
    return ('G');
  case 3:
    return ('U');
  default:
    perror("RNAi2c: I got a non RNA interger");
    return ('\0');
  }
}

/* int RNAc2i(char rnachar)
   Parameter: give a RNA character (ACGU).
   Description: RNA character to RNA integer
   Return value: 0 for 'A', 1 for 'C', 2 for 'G', 3 for 'U', -1 for error..
*/
int RNAc2i(char rnachar)
{
  switch (rnachar) {
  case 'A':
    return 0;
  case 'C':
    return 1;
  case 'G':
    return 2;
  case 'U':
    return 3;
  default:
    perror("RNAc2i: I got a non RNA character");
    return -1;
  }
}

/********************************
 *          Inputting           *
 ********************************/

int Args(int argc, char **argv)
{
  int i;
  int It;
  char	*Ct;
  /* parameters may be overridden here. */
  for (i = 1; i < argc; i++) {
    if (strcmp(argv[i],"-Distance") == 0){
      i++; if (i == argc) return 1;
      It = atoi(argv[i]);
      if (It==1)
	distanceG=1;
      else if (It==2)
	distanceG=2;
      else {
	fprintf(stderr,"Args: wrong value for -Distance\n");
	return 1;
      }
      continue;
    }
    else if (strcmp(argv[i],"-InitialSeq") == 0){
      i++; if(i==argc) return 1;
      Ct = argv[i];
      if ((strlen(Ct) > 0) && (strlen(Ct) <= MAXLEN-1)) 
	strcpy(initSeqG, Ct);
      else {
	fprintf(stderr,"Args: sequence length must be less than %d\n",MAXLEN);
	return 1;
      }
      continue;
    }
    else if (strcmp(argv[i],"-TargetStr") == 0){
      i++; if(i==argc) return 1;
      Ct = argv[i];
      if ((strlen(Ct) > 0) && (strlen(Ct) <= MAXLEN-1)) 
	strcpy(targetStrG, Ct);
      else {
	fprintf(stderr,"Args: target length must be less than %d\n",MAXLEN);
	return 1;
      }
      continue;
    }
    else {
      fprintf(stderr,"-Distance default=%d\n",distanceG);
      fprintf(stderr,"-InitialSeq (obligatory)\n");
      fprintf(stderr,"-TargetStr (obligatory)\n");
      exit(1);
    }
  }
  return (0);
}


